<?php

declare(strict_types=1);

namespace Drupal\content_commander;

use Drupal\content_commander\Command\CcCommand;
use Symfony\Component\Console\Attribute\AsCommand;
use Symfony\Component\DependencyInjection\Compiler\CompilerPassInterface;
use Symfony\Component\DependencyInjection\ContainerBuilder;
use Symfony\Component\DependencyInjection\Definition;

/**
 * Console Commander compiler pass.
 *
 * Inspired by:
 *  - \Drupal\dex_console\DexCompilerPass()
 *  - \Drupal\pinto\PintoCompilerPass()
 */
final class CcCompilerPass implements CompilerPassInterface {

  public function process(ContainerBuilder $container): void {
    /** @var array<class-string, string> $containerNamespaces */
    $containerNamespaces = $container->getParameter('container.namespaces');

    /** @var string[] $ccNamespaces */
    $ccNamespaces = $container->getParameter('content_commander.namespaces');

    $all = [];
    foreach ($this->getEnums($containerNamespaces, $ccNamespaces) as $r) {
      /** @var class-string<\Drupal\content_commander\ContentInterface> $className */
      $className = $all[] = $r->getName();

      // Don't create a service definition if this class is already a service.
      if ($container->hasDefinition($className) || $container->hasAlias($className)) {
        continue;
      }

      /** @var \Symfony\Component\Console\Attribute\AsCommand $attribute */
      $attribute = ($r->getAttributes(AsCommand::class)[0] ?? throw new \LogicException('Not possible due to getEnums checks'))
        ->newInstance();

      $definition = new Definition(CcCommand::class);
      $definition
        ->setAutoconfigured(TRUE)
        ->setAutowired(TRUE)
        ->setPublic(TRUE)
        ->setArgument('$name', $attribute->name)
        ->setArgument('$ccEnums', [$className])
        ->setTags(['console.command' => [['command' => $attribute->name] + \get_object_vars($attribute)]]);

      $container->setDefinition($className, $definition);
    }

    if ($all !== []) {
      $definition = new Definition(CcCommand::class);
      $definition
        ->setAutoconfigured(TRUE)
        ->setAutowired(TRUE)
        ->setPublic(TRUE)
        ->setArgument('$name', 'content-commander:create-all')
        ->setArgument('$ccEnums', $all)
        ->setTags([
          'console.command' => [
            [
              'command' => 'content-commander:create-all',
            ],
          ],
        ]);

      $container->setDefinition('.content_commander.command.all', $definition);
    }
  }

  /**
   * Get enums for the provided namespaces.
   *
   * @param array<class-string, string> $namespaces
   *   An array of namespaces. Where keys are class strings and values are
   *   paths.
   * @param string[] $ccNamespaces
   *   Content Commander namespaces.
   *
   * @return \Generator<\ReflectionClass<\Drupal\content_commander\ContentInterface>>
   *   Generates class strings.
   *
   * @throws \ReflectionException
   */
  private function getEnums(array $namespaces, array $ccNamespaces): \Generator {
    foreach ($namespaces as $namespace => $dirs) {
      $dirs = (array) $dirs;
      foreach ($dirs as $dir) {
        foreach ($ccNamespaces as $ccNamespace) {
          $nsDir = $dir . '/' . \str_replace('\\', '/', $ccNamespace);
          if (!\file_exists($nsDir)) {
            continue;
          }
          $namespace .= '\\' . $ccNamespace;

          /** @var \RecursiveIteratorIterator<\RecursiveDirectoryIterator<\SplFileInfo>> $iterator */
          $iterator = new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($nsDir, \FilesystemIterator::KEY_AS_PATHNAME | \FilesystemIterator::CURRENT_AS_FILEINFO | \RecursiveDirectoryIterator::SKIP_DOTS), \RecursiveIteratorIterator::SELF_FIRST);
          foreach ($iterator as $fileinfo) {
            \assert($fileinfo instanceof \SplFileInfo);
            if ('php' !== $fileinfo->getExtension()) {
              continue;
            }

            /** @var \RecursiveDirectoryIterator|null $subDir */
            $subDir = $iterator->getSubIterator();
            if (NULL === $subDir) {
              continue;
            }

            $subDir = $subDir->getSubPath();
            $subDir = $subDir !== '' ? \str_replace(DIRECTORY_SEPARATOR, '\\', $subDir) . '\\' : '';

            /** @var class-string $class */
            $class = $namespace . '\\' . $subDir . $fileinfo->getBasename('.php');

            try {
              /** @var \ReflectionClass<\Drupal\content_commander\ContentInterface> $reflectionClass */
              $reflectionClass = new \ReflectionClass($class);
            }
            // @phpstan-ignore-next-line
            catch (\Error) {
              // Skip commands where the hierarchy is unresolvable due to
              // optional dependencies.
              continue;
            }

            if (FALSE === $reflectionClass->isEnum() || FALSE === $reflectionClass->implementsInterface(ContentInterface::class)) {
              continue;
            }

            if (\count($reflectionClass->getAttributes(AsCommand::class)) > 0) {
              yield $reflectionClass;
            }
          }
        }
      }
    }
  }

}
