<?php

declare(strict_types=1);

namespace Drupal\content_commander\Command;

use Digilist\DependencyGraph\CircularDependencyException;
use Digilist\DependencyGraph\DependencyGraph;
use Digilist\DependencyGraph\DependencyNode;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\content_commander\Attribute\DependsOn;
use Drupal\content_commander\ContentContext;
use Drupal\content_commander\ContentInterface;
use Drupal\content_commander\Utils\Utils;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityTypeRepositoryInterface;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;

/**
 * This command is manually registered to the container.
 */
final class CcCommand extends Command {

  private const IDE_LAUNCH = [
    'emacs' => 'emacs://open?url=file://%s&line=%s',
    'macvim' => 'mvim://open?url=file://%s&line=%s',
    'phpstorm' => 'phpstorm://open?file=%s&line=%s',
    'sublime' => 'subl://open?url=file://%s&line=%s',
    'textmate' => 'txmt://open?url=file://%s&line=%s',
    'vscode' => 'vscode://file/%s:%s',
  ];

  private string $ideLaunch = self::IDE_LAUNCH['phpstorm'];
  private ?string $nativeFilePrefix = NULL;

  /**
   * @phpstan-param array<class-string<\Drupal\content_commander\ContentInterface>> $ccEnums
   */
  public function __construct(
    string $name,
    private readonly array $ccEnums,
    private readonly EntityTypeRepositoryInterface $entityTypeRepository,
    private readonly EntityRepositoryInterface $entityRepository,
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly ConfigFactoryInterface $configFactory,
    private readonly UuidInterface $uuid,
  ) {
    parent::__construct($name);
  }

  protected function configure(): void {
    $this->addOption('delete', 'd', InputOption::VALUE_NONE);
  }

  protected function execute(InputInterface $input, OutputInterface $output): int {
    $io = new SymfonyStyle($input, $output);

    $env = \getenv('NATIVE_FILE_PREFIX');
    $this->nativeFilePrefix = $env !== FALSE ? $env : NULL;

    ($isDeleting = (bool) $input->getOption('delete'))
      ? $io->warning('Deleting existing content')
      : $io->note('Not deleting existing content');

    foreach ($this->ccEnums as $enum) {
      $code = $this->process($io, $enum, $isDeleting);
      if ($code !== static::SUCCESS) {
        return $code;
      }
    }

    $io->success('Done.');

    return static::SUCCESS;
  }

  /**
   * @phpstan-param class-string<\Drupal\content_commander\ContentInterface> $ccEnum
   */
  private function process(SymfonyStyle $io, string $ccEnum, bool $isDeleting): int {
    $rClass = new \ReflectionClass($ccEnum);
    $classFileName = $rClass->getFileName();
    if ($classFileName === FALSE) {
      throw new \LogicException();
    }

    $io
      ->section(\sprintf(
        "<href=%s>%s</>",
        $this->ideLink($classFileName, (int) $rClass->getStartLine()),
        $ccEnum,
      ));

    /** @var \SplObjectStorage<\Drupal\content_commander\ContentInterface, DependencyNode<\Drupal\content_commander\ContentInterface>> $dependencyNodes */
    $dependencyNodes = new \SplObjectStorage();
    foreach ($ccEnum::cases() as $case) {
      ($dependencyNodes[$case] = new DependencyNode($case));
    }

    /** @var \SplObjectStorage<\Drupal\content_commander\ContentInterface, \SplObjectStorage<\Drupal\content_commander\ContentInterface, mixed>> $requireUsageDependencies */
    $requireUsageDependencies = new \SplObjectStorage();
    foreach ($dependencyNodes as $enum) {
      /** @var \SplObjectStorage<\Drupal\content_commander\ContentInterface, mixed> $required */
      $required = new \SplObjectStorage();
      $r = new \ReflectionClassConstant($enum::class, $enum->name);
      foreach ($r->getAttributes(DependsOn::class) as $dependency) {
        /** @var \Drupal\content_commander\Attribute\DependsOn $dependsOn */
        $dependsOn = $dependency->newInstance();
        $dependencyNodes[$enum]->dependsOn($dependencyNodes[$dependsOn->content]);
        if ($dependsOn->optional === FALSE) {
          $required->attach($dependsOn->content);
        }
      }
      $requireUsageDependencies[$enum] = $required;
    }

    $graph = new DependencyGraph();
    foreach ($dependencyNodes as $enum) {
      // addNode must be called only after dependencies have been added!
      $graph->addNode($dependencyNodes[$enum]);
    }

    try {
      /** @var \Drupal\content_commander\ContentInterface[] $resolved */
      $resolved = $graph->resolve();
    }
    catch (CircularDependencyException $e) {
      $io->error($e->getMessage());
      return static::FAILURE;
    }

    // Map of entity type ID to UUID key.
    /** @var array<string, string> $uuidKeys */
    $uuidKeys = [];

    $metricsEntitiesDeleted = 0;
    $metricsEntitiesSkipped = 0;
    $metricsEntitiesCreated = 0;

    /** @var \SplObjectStorage<\Drupal\content_commander\ContentInterface, \Drupal\Core\Entity\ContentEntityInterface> $entities */
    $entities = new \SplObjectStorage();
    foreach ($resolved as $enum) {
      $uuid = $enum->uuid();
      if ($uuid === NULL) {
        /** @var array<string, string> $map */
        $map = $this->configFactory
          ->getEditable('content_commander.content_mapping')
          ->get('map') ?? [];

        $enumKey = $enum::class . '::' . $enum->name;
        $uuid = \array_search($enumKey, $map, TRUE);
        if ($uuid === FALSE) {
          // Generate and store a UUID in configuration.
          $uuid = $this->uuid->generate();
          $map[$uuid] = $enumKey;
          $this->configFactory->getEditable('content_commander.content_mapping')
            ->set('map', $map)
            ->save(TRUE);
        }
      }

      $entityClass = $enum->entityClass();
      $entityTypeId = $this->entityTypeRepository->getEntityTypeFromClass($entityClass);

      if ($io->isDebug()) {
        $io->info(\sprintf('Starting Entity %s %s.', $entityTypeId, $uuid));
      }

      /** @var \Drupal\Core\Entity\ContentEntityInterface|null $entity */
      $entity = $this->entityRepository->loadEntityByUuid($entityTypeId, $uuid);

      // If the entity exists, and we're not deleting, skip.
      if (!$isDeleting && $entity !== NULL) {
        $metricsEntitiesSkipped++;
        $entities[$enum] = $entity;

        if ($io->isDebug()) {
          $io->info(\sprintf('Entity %s %s already exists and will not be deleted.', $entityTypeId, $uuid));
        }

        continue;
      }

      // If the entity exists, and we're deleting, delete.
      if ($isDeleting && $entity !== NULL) {
        $metricsEntitiesDeleted++;
        $entity->delete();
        if ($io->isDebug()) {
          $io->warning(\sprintf('Entity %s %s already exists and was deleted.', $entityTypeId, $uuid));
        }
      }

      if ($io->isDebug()) {
        $io->info(\sprintf('Entity %s %s will be recreated.', $entityTypeId, $uuid));
      }

      // Create a fresh entity.
      $metricsEntitiesCreated++;
      $entities[$enum] = $newEntity = $entityClass::create();
      // Ignore faulty typehinting (exception, not null).
      // @phpstan-ignore-next-line
      $uuidKeys[$entityTypeId] ??= $this->entityTypeManager->getDefinition($entityTypeId)->getKey('uuid') ?: throw new \Exception(\sprintf('%s entities must support UUIDs', $entityClass));
      $newEntity->set($uuidKeys[$entityTypeId], $uuid);

      // Determine the entity dependencies for this enum.
      /** @var \SplObjectStorage<\Drupal\content_commander\ContentInterface, \Drupal\Core\Entity\ContentEntityInterface> $entityDependencies */
      $entityDependencies = new \SplObjectStorage();
      foreach ($dependencyNodes[$enum]->getDependencies() as $dependency) {
        $entityDependencies[$dependency->getElement()] = $entities[$dependency->getElement()];
      }
      $contentContext = ContentContext::create(
        for: $enum,
        dependencies: $entityDependencies,
        requiredDependencies: $requireUsageDependencies[$enum],
      );
      $enum->createContent(contentContext: $contentContext, entity: $newEntity);
      $unusedDependencies = \iterator_to_array($contentContext->unusedDependencies());
      if (\count($unusedDependencies) !== 0) {
        [$fileName, $line] = Utils::locationFor($enum);

        $io->error('Unused entity dependency');
        $io->text(
          \sprintf(
            "<href=%s>%s</> failed to use the entity represented by %s. Use the `optional` parameter on %s if usage is intentionally conditional.",
            $this->ideLink($fileName, $line),
            ($enum::class . '::' . $enum->name),
            \implode(
              ', ',
              \array_map(
                static fn (ContentInterface $enum): string => ($enum::class . '::' . $enum->name),
                $unusedDependencies,
              ),
            ),
            DependsOn::class,
          ));
        $io->newLine();

        return static::FAILURE;
      }

      $newEntity->save();

      if ($io->isVerbose()) {
        $io->success(\sprintf('Entity %s %s was saved.', $entityTypeId, $uuid));
      }
    }

    $io->info(\sprintf(
      '%d entities were created. %d existing entities were deleted. %d existing entities were skipped.',
      $metricsEntitiesCreated,
      $metricsEntitiesDeleted,
      $metricsEntitiesSkipped,
    ));

    return static::SUCCESS;
  }

  private function ideLink(string $path, int $line): string {
    $dir = \getcwd();
    if ($dir === FALSE) {
      throw new \Exception('Unable to determine CWD');
    }

    $path = \str_starts_with($path, $dir) ? \substr($path, \strlen($dir)) : $path;

    $path = (NULL !== $this->nativeFilePrefix)
      ? $this->nativeFilePrefix . $path
      : $path;

    return \sprintf($this->ideLaunch, $path, (string) $line);
  }

}
