<?php

declare(strict_types=1);

namespace Drupal\content_commander\Utils;

use Drupal\content_commander\ContentInterface;
use PhpParser\Node;
use PhpParser\Node\Stmt\Enum_;
use PhpParser\Node\Stmt\EnumCase;
use PhpParser\NodeTraverser;
use PhpParser\NodeVisitorAbstract;
use PhpParser\ParserFactory;

/**
 * Internal utilities.
 *
 * These may change/remove at any time.
 *
 * @internal
 */
final class Utils {

  /**
   * @return array{string, int}
   *   File and line number where an enum case is located in code.
   *
   * @throws \Error
   *   When enum file cannot be parsed.
   */
  public static function locationFor(ContentInterface $enum): array {
    $rClass = new \ReflectionClass($enum::class);

    $propertyVisitor = new class ($rClass->getShortName(), $enum->name) extends NodeVisitorAbstract {

      /**
       * The line number, if enum case is found.
       *
       * @var int|null
       */
      public ?int $lineNumber = NULL;

      public function __construct(
        private string $classShortName,
        private string $caseName,
      ) {
      }

      public function enterNode(Node $node): NULL {
        if ($node instanceof Enum_ && $node->name?->name === $this->classShortName) {
          foreach ($node->stmts as $stmt) {
            if ($stmt instanceof EnumCase) {
              if ($this->caseName === $stmt->name->name) {
                $this->lineNumber = $stmt->getStartLine();
              }
            }
          }
        }

        return NULL;
      }

    };

    $fileName = $rClass->getFileName();
    if ($fileName === FALSE) {
      throw new \LogicException();
    }

    $classFile = \file_get_contents($fileName);
    if ($classFile === FALSE) {
      throw new \LogicException();
    }

    $parser = (new ParserFactory())->createForNewestSupportedVersion();
    $stmts = $parser->parse($classFile) ?? throw new \LogicException();

    $traverser = new NodeTraverser();
    $traverser->addVisitor($propertyVisitor);
    $traverser->traverse($stmts);

    if ($propertyVisitor->lineNumber !== NULL) {
      return [$fileName, $propertyVisitor->lineNumber];
    }

    throw new \LogicException('Could not find case data for ' . $enum->name);
  }

}
